import base64
import os
from pathlib import Path
from typing import Dict, List, Union

import plotly.graph_objects as go

from fiat_toolbox.infographics.infographics import InfographicsParser
from fiat_toolbox.infographics.infographics_interface import IInfographicsParser
from fiat_toolbox.metrics_writer.fiat_read_metrics_file import (
    MetricsFileReader,
)


class RiskInfographicsParser(IInfographicsParser):
    """Class for creating the infographic"""

    def __init__(
        self,
        scenario_name: str,
        metrics_full_path: Union[Path, str],
        config_base_path: Union[Path, str],
        output_base_path: Union[Path, str],
    ) -> None:
        """Initialize the InfographicsParser

        Parameters
        ----------
        scenario_name : str
            The name of the scenario
        metrics_full_path : Union[Path, str]
            The path to the metrics file
        config_base_path : Union[Path, str]
            The path to the config folder
        output_base_path : Union[Path, str]
            The path to the output folder
        """

        # Save the scenario name
        self.scenario_name = scenario_name

        # Convert the metrics path to a Path object
        if isinstance(metrics_full_path, str):
            metrics_full_path = Path(metrics_full_path)
        self.metrics_full_path = metrics_full_path

        # Convert the config path to a Path object
        if isinstance(config_base_path, str):
            config_base_path = Path(config_base_path)
        self.config_base_path = config_base_path

        # Convert the output path to a Path object
        if isinstance(output_base_path, str):
            output_base_path = Path(output_base_path)
        self.output_base_path = output_base_path

    def _get_impact_metrics(self) -> Dict:
        """Get the impact metrics for a scenario

        Returns
        -------
        Dict
            The impact metrics for the scenario
        """

        # Check if the metrics file exists
        if not Path.exists(self.metrics_full_path):
            raise FileNotFoundError(
                f"Metrics file not found at {self.metrics_full_path}"
            )

        # Read configured metrics
        metrics = (
            MetricsFileReader(self.metrics_full_path)
            .read_metrics_from_file()
            .to_dict()["Value"]
        )

        # Return the metrics
        return metrics

    @staticmethod
    def _encode_image_from_path(image_path: str) -> str:
        """Encode an image from a path to a base64 string

        Parameters
        ----------
        image_path : str
            The path to the image

        Returns
        -------
        str
            The base64 encoded image string
        """
        if os.path.isfile(image_path):
            with open(image_path, "rb") as image_file:
                encoded_string = base64.b64encode(image_file.read()).decode()
            return f'data:image/png;base64,{encoded_string}'

    @staticmethod
    def _figures_list_to_html(
        rp_fig: go.Figure,
        metrics: Dict,
        charts: Dict,
        file_path: Union[str, Path] = "infographics.html",
        image_path: Union[str, Path] = None,
    ):
        """Save a list of plotly figures in an HTML file

        Parameters
        ----------
            rp_fig : go.Figure
                The plotly figure consisting of the pie charts for multiple return periods
            metrics : Dict
                The impact metrics for the scenario
            file_path : Union[str, Path], optional
                Path to the HTML file, by default "infographics.html"
            image_path : Union[str, Path], optional
                Path to the image folder, by default None
        """

        # Convert the file_path to a Path object
        if isinstance(file_path, str):
            file_path = Path(file_path)

        # Check if the file_path already exists
        if Path.exists(file_path):
            raise FileExistsError(f"File already exists at {file_path}")

        # Check if the file_path is correct
        if file_path.suffix != ".html":
            raise ValueError(f"File path must be a .html file, not {file_path}")

        # Create the directory if it does not exist
        if not Path.exists(file_path.parent):
            file_path.parent.mkdir(parents=True)

        # Check if the image_path exists
        expected_damage_path = InfographicsParser._check_image_source(charts['Other']['Expected_Damages']['image'], image_path, return_image=False)
        flooded_path = InfographicsParser._check_image_source(charts['Other']['Flooded']['image'], image_path, return_image=False)

        # Div height is the max of the chart heights
        div_height = max(charts['Other']['Expected_Damages']['height'], charts['Other']['Flooded']['height'], charts['Other']['Return_Periods']['plot_height'])

        # Write the html to the file
        with open(file_path, "w", encoding="utf-8") as infographics:
            rp_charts = rp_fig.to_html(config={'displayModeBar': False}).split("<body>")[1].split("</body>")[0]

            infographics.write(
                f"""
                <!DOCTYPE html>
                <html>
                <head>
                    <title></title>
                    <style>
                        .container {{
                            display: flex;
                            flex-direction: column;
                            align-items: center;
                            justify-content: center;  # Center the plots vertically
                            height: 100vh;
                        }}
                        .inner-div {{
                            text-align: center;
                            max-height: {div_height}px; /* Add your max height here */
                            overflow: auto; /* Add this to handle content that exceeds the max height */
                            display: flex;
                            flex-direction: column;
                            align-items: center;
                        }}
                        .chart-container {{
                            /* Add your CSS styling for chart container here */
                        }}
                        .img-container1 {{
                            max-width: 10%;
                            height: auto;
                            margin: 0 auto;
                            transform: scale({charts['Other']['Expected_Damages']['image_scale']}); /* Add your scale factor here */
                        }}
                        .img-container2 {{
                            max-width: 10%;
                            height: auto;
                            margin: 0 auto;
                            transform: scale({charts['Other']['Flooded']['image_scale']}); /* Add your scale factor here */
                        }}
                        h1 {{
                            font-size: {charts['Other']['Expected_Damages']['title_font_size']}px; /* Adjust the font size as needed */
                            font-family: Verdana; /* Specify the font family as Verdana */
                            font-weight:normal; 
                        }}
                        h2 {{
                            font-size: {charts['Other']['Flooded']['title_font_size']}px; /* Adjust the font size as needed */
                            font-family: Verdana; /* Specify the font family as Verdana */
                            font-weight:normal; 
                        }}
                        p1 {{
                            font-size: {charts['Other']['Flooded']['numbers_font_size']}px; /* Adjust the font size as needed */
                            font-family: Verdana; /* Specify the font family as Verdana */
                        }}
                        p2 {{
                            font-size: {charts['Other']['Flooded']['numbers_font_size']}px; /* Adjust the font size as needed */
                            font-family: Verdana; /* Specify the font family as Verdana */
                        }}
                    </style>
                </head>
                <body>
                    <div class="container">
                        <div class="inner-div">
                            <h1>{charts['Other']['Expected_Damages']['title']}</h1>
                            <img src="{RiskInfographicsParser._encode_image_from_path(expected_damage_path)}" alt="Expected Damage" class="img-container1">
                            <p1>${'{:,.0f}'.format(metrics['ExpectedAnnualDamages'])}</p1>
                        </div>
                        <div class="inner-div">
                            <h2>{charts['Other']['Flooded']['title']}</h2>
                            <img src="{RiskInfographicsParser._encode_image_from_path(flooded_path)}" alt="Flooded Homes" class="img-container2">
                            <p2>{'{:,.0f}'.format(metrics['FloodedHomes'])}</p2>
                        </div>
                        <div class="inner-div chart-container">
                            {rp_charts}
                        </div>
                    </div>
                </body>
                </html>
                """
            )

    def _get_infographics(
        self,
    ) -> Union[Dict, Dict, go.Figure]:
        """Get the infographic for a scenario

        Returns
        -------
        go.Figure
            The infographic for the scenario

        """

        # Get the impact metrics
        metrics = self._get_impact_metrics()

        # Get the infographic configuration
        pie_chart_config_path = self.config_base_path.joinpath(
            "config_risk_charts.toml"
        )

        # Check if the infographic configuration files exist
        if not Path.exists(pie_chart_config_path):
            raise FileNotFoundError(
                f"Infographic configuration file not found at {pie_chart_config_path}"
            )

        # Get the pie chart dictionaries
        charts = InfographicsParser._get_pies_dictionary(pie_chart_config_path, metrics)

        # Create the pie chart figures
        charts_fig = InfographicsParser._get_pie_chart_figure(
            data=charts.copy(),
            legend_orientation="h",
            yanchor="top",
            y=-0.1,
            title=charts['Other']['Return_Periods']['title'],
            image_path=self.config_base_path.joinpath("images"),
            title_font_size=charts['Other']['Return_Periods']['font_size'],
            subtitle_font_size=charts['Other']['Return_Periods']['subtitle_font'],
            image_scale=charts['Other']['Return_Periods']['image_scale'],
            numbers_font=charts['Other']['Return_Periods']['numbers_font'],
            legend_font_size=charts['Other']['Return_Periods']['legend_font'],
            plot_info=charts["Other"]['Info']["text"],
            plot_info_img=charts["Other"]["Info"]["image"],
            plot_info_scale=charts["Other"]["Info"]["scale"],
            plot_height=charts['Other']['Return_Periods']['plot_height'],
        )

        # Return the figure
        return metrics, charts, charts_fig

    def get_infographics(self) -> Union[List[go.Figure], go.Figure]:
        """Get the infographic for a scenario

        Returns
        -------
        Union[List[go.Figure], go.Figure]
            The infographic for the scenario as a list of figures or a single figure
        """

        # Get the infographic
        _, _, infographic = self._get_infographics()

        # Return the infographic
        return infographic

    def write_infographics_to_file(self) -> str:
        """Write the infographic for a scenario to file

        Returns
        -------
        str
            The path to the infographic file
        """

        # Create the infographic path
        infographic_html = self.output_base_path.joinpath(
            f"{self.scenario_name}_metrics.html"
        )

        # Check if the infographic already exists. If so, return the path
        if Path.exists(infographic_html):
            # TODO: Print logging message
            return str(infographic_html)

        # Get the infographic
        metrics, charts, infographic = self._get_infographics()

        # Convert the infographic to html. The default for using relative image paths is to have an images folder in the same directory as the config files
        self._figures_list_to_html(infographic, metrics, charts, infographic_html, self.config_base_path.joinpath("images"))

        # Return the path to the infographic
        return str(infographic_html)

    def get_infographics_html(self) -> str:
        """Get the path to the infographic html file

        Returns
        -------
        str
            The path to the infographic html file
        """

        # Create the infographic path
        infographic_path = self.output_base_path.joinpath(
            f"{self.scenario_name}_metrics.html"
        )

        return str(infographic_path)



import os
from pathlib import Path
from typing import Union

import pandas as pd

from fiat_toolbox.metrics_writer.fiat_metrics_interface import IMetricsFileReader


class MetricsFileReader(IMetricsFileReader):
    """Reads metrics from a file."""

    def __init__(self, metrics_file_path: Union[str, Path]):
        """
        Initializes a new instance of the MetricsFileReader class.

        Parameters
        ----------
        metrics_file_path : str
            The path to the file containing the metrics.

        Raises
        ------
        FileNotFoundError
            If the file cannot be found.
        ValueError
            If the file is not a valid metrics file.
        """

        # Convert the path to a Path object
        if not isinstance(metrics_file_path, Path):
            metrics_file_path = Path(metrics_file_path)

        # Check if the file is a csv file
        if not metrics_file_path.suffix == ".csv":
            raise ValueError("The file must be a csv file.")

        # Check if the file exists
        if not os.path.exists(metrics_file_path):
            raise FileNotFoundError("The file does not exist.")

        # Set the metrics file path
        self.metrics_file_path = metrics_file_path

    def read_aggregated_metric_from_file(self, metric: str) -> pd.Series:
        """Reads metrics from a file. These metrics are aggregated metrics.

        Parameters:
        ----------
        metric: str
            The metric to read from the file.

        Returns:
        -------
        pd.DataFrame
            The metrics read from the file.

        Raises:
        ------
        KeyError
            If the metric is not found in the file.
        """

        # Read the metrics from the file
        df_metrics = pd.read_csv(self.metrics_file_path, index_col=0)

        # Remove the desctioption row
        df_metrics = df_metrics.iloc[1:]

        # Check if the metric is in the dataframe
        if metric not in df_metrics.columns:
            raise KeyError(f"The metric {metric} was not found in the file.")

        # Return the metric
        return df_metrics[metric]

    def read_metrics_from_file(self, **kwargs) -> pd.Series:
        """
        Reads metrics from a file.

        Parameters
        ----------
        include_long_names : bool
            Include the long names of the metrics.
        include_metrics_table_selection : bool
            Include the metrics table selection.
        include_description : bool
            Include the description of the metrics.

        Returns
        -------
        pd.DataFrame
            The metrics read from the file.

        Raises
        ------
        KeyError
            If the metric is not found in the file.
        """

        # Set the default values
        include_long_names = kwargs.get("include_long_names", False)
        include_metrics_table_selection = kwargs.get(
            "include_metrics_table_selection", False
        )
        include_description = kwargs.get("include_description", False)

        # Read the metrics from the file
        df_metrics = pd.read_csv(self.metrics_file_path, index_col=0)

        # If you can't grab the value, transpose the data
        if 'Value' not in df_metrics.columns:
            df_metrics = df_metrics.transpose()
            
        # Ensure values are interpreted as numbers
        df_metrics["Value"] = pd.to_numeric(df_metrics["Value"])

        # Remove the desctioption row
        if not include_description:
            df_metrics = df_metrics.drop("Description", axis="columns")

        # Remove the long names row
        if not include_long_names:
            df_metrics = df_metrics.drop("Long Name", axis="columns")

        # Remove the metrics table selection row
        if not include_metrics_table_selection:
            df_metrics = df_metrics.drop("Show In Metrics Table", axis="columns")

        # Return the metric
        return df_metrics



from abc import ABC, abstractmethod
from pathlib import Path

import pandas as pd


class IMetricsFileWriter(ABC):
    """Interface for writing metrics to a file."""

    @abstractmethod
    def parse_metrics_to_file(
        self,
        df_results: pd.DataFrame,
        metrics_path: Path,
        write_aggregate: str = None,
        overwrite: bool = False,
    ) -> None:
        """
        Parse a metrics file and write the metrics to a file.

        Parameters
        ----------
        df_results : pd.DataFrame
            The results dataframe.
        metrics_path : Path
            The path to where to store the metrics file.
        write_aggregate : str
            The name of the aggregation label to write to the metrics file
            (None for no aggregation label, 'all' for all possible ones).
        overwrite : bool
            Whether to overwrite the existing metrics file if it already exists.
        """
        pass


class IMetricsFileReader(ABC):
    """Interface for reading metrics from a file."""

    @abstractmethod
    def read_metrics_from_file(self, **kwargs) -> pd.Series:
        """
        Reads metrics from a file.

        Parameters
        ----------
        **kwargs
            Keyword arguments.

        Returns
        -------
        pd.DataFrame
            The metrics read from the file.

        Raises
        ------
        KeyError
            If the metric is not found in the file.
        """

        pass

    @abstractmethod
    def read_aggregated_metric_from_file(self, metric: str) -> pd.Series:
        """
        Reads metrics from a file. These metrics are aggregated metrics.

        Parameters
        ----------
        metric : str
            The metric to read from the file.

        Returns
        -------
        pd.DataFrame
            The metrics read from the file.

        Raises
        ------
        KeyError
            If the metric is not found in the file.
        """

        pass




import numpy as np
import pandas as pd


class ExceedanceProbabilityCalculator:
    def __init__(self, column_prefix):
        self.column_prefix = column_prefix

    def append_probability(self, df: pd.DataFrame, threshold: float, T: float) -> pd.DataFrame:
        """Append exceedance probability to dataframe.
        
        Parameters
        ----------
        df : pandas.DataFrame
            Dataframe containing the data.
        threshold : float
            Threshold value.
        T : float
            Time horizon.
        
        Returns
        -------
        pandas.DataFrame
            Dataframe containing the data and the exceedance probability.
        """
        
        # Initialize result dataframe
        result = df.copy()

        # Calculate exceedance probability
        result['Exceedance Probability'] = self.calculate(df, threshold, T)

        return result

    def calculate(self, df: pd.DataFrame, threshold: float, T: float) -> pd.DataFrame:
        """Calculate exceedance probability.

        Parameters
        ----------
        df : pandas.DataFrame
            Dataframe containing the data.
        threshold : float
            Threshold value.
        T : float
            Time horizon.

        Returns
        -------
        pandas.DataFrame    
            Dataframe containing the exceedance probability.
        """

        # Extract return periods from column names
        return_periods = [int(col.split('(')[1][:-2]) for col in df.columns if col.startswith(self.column_prefix)]
        
        # Calculate exceedance probability
        return self._calculate(df, return_periods, threshold, T).to_frame()
        
    def append_to_file(self, input_file: str, output_file: str, threshold: float, T: float) -> None:
        """Append exceedance probability to file.

        Parameters
        ----------
        input_file : str
            Path to input file.
        output_file : str
            Path to output file.
        threshold : float
            Threshold value.
        T : float
            Time horizon.
        """

        # Read data from file
        df = pd.read_csv(input_file, index_col=0)

        # Append exceedance probability
        result = self.append_probability(df, threshold, T)

        # Write data to file
        result.to_csv(output_file)

    def _calculate(self, df: pd.DataFrame, return_periods: list, threshold: float, T: float) -> pd.Series:
        """Calculate exceedance probability.

        Parameters
        ----------
        df : pandas.DataFrame
            Dataframe containing the data.
        return_periods : list
            List of return periods.
        threshold : float
            Threshold value.
        T : float
            Time horizon.

        Returns
        -------
        pandas.Series
            Series containing the exceedance probability.
        """

        # Convert all non-numerical values to nan
        df = df.apply(lambda x: pd.to_numeric(x, errors='coerce'))

        # Extract values for the selected columns
        values = df.filter(like=self.column_prefix).to_numpy()

        # Create a mask where True indicates a NaN value
        nan_mask = np.isnan(values)

        # Check if there are any NaN values after the first non-NaN value in each row
        invalid_rows = np.any(np.diff(nan_mask.astype(int), axis=1) == 1, axis=1)

        # Add the check if all elements in a row are NaN
        invalid_rows = invalid_rows | np.all(nan_mask, axis=1)

        # Custom interpolation function
        def custom_interp(x, xp, fp):
            if x > xp[-1]:
                return np.nan
            elif x < xp[0]:
                return fp[0]
            else: 
                return np.interp(x, xp, fp)

        # Interpolate to find the return period for which the threshold is first exceeded
        RP = np.array([custom_interp(threshold, row, return_periods) for row in values])

        # Calculate exceedance probability
        mask = ~invalid_rows
        result = np.full(len(df), np.nan)
        result[mask] = np.round((1 - np.exp(-T / RP[mask])) * 100, 1)

        return pd.Series(result, name = 'Exceedance Probability')


from pathlib import Path
from typing import Union

from fiat_toolbox.infographics.infographics import InfographicsParser
from fiat_toolbox.infographics.infographics_interface import IInfographicsParser
from fiat_toolbox.infographics.risk_infographics import RiskInfographicsParser


class InforgraphicFactory:
    @staticmethod
    def create_infographic_file_writer(
        infographic_mode: str,
        scenario_name: str,
        metrics_full_path: Union[Path, str],
        config_base_path: Union[Path, str],
        output_base_path: Union[Path, str],
    ) -> IInfographicsParser:
        """
        Create a infographic file writer.

        Parameters
        ----------
        infographic_mode : str
            The mode of the infographic file writer to create.
        config_file : Path
            The path to the infographic file.

        Returns
        -------
        IInfographicsFileWriter
            A infographic file writer.
        """
        if infographic_mode == "single_event":
            return InfographicsParser(
                scenario_name, metrics_full_path, config_base_path, output_base_path
            )
        elif infographic_mode == "risk":
            return RiskInfographicsParser(
                scenario_name, metrics_full_path, config_base_path, output_base_path
            )
        else:
            raise ValueError(f"Infographic_mode {infographic_mode} not supported")



from abc import ABC, abstractmethod
from collections import Counter
from pathlib import Path
from typing import Optional, Union

import geopandas as gpd
import numpy as np
import pandas as pd


class IPointsToFootprints(ABC):
    """Interface for writing a footprints spatial file."""

    @abstractmethod
    def write_footprint_file(
        footprints: gpd.GeoDataFrame, results: pd.DataFrame
    ) -> None:
        pass


class PointsToFootprints(IPointsToFootprints):
    """Write a footprints spatial file."""

    @staticmethod
    def _check_extension(out_path, ext):
        out_path = Path(out_path)
        if out_path.suffix != ext:
            raise ValueError(
                f"File extention given: '{out_path.suffix}' does not much the file format specified: {ext}."
            )

    @staticmethod
    def _mode(my_list):
        ct = Counter(my_list)
        max_value = max(ct.values())
        return sorted(key for key, value in ct.items() if value == max_value)

    @staticmethod
    def write_footprint_file(
        footprints: gpd.GeoDataFrame,
        points: pd.DataFrame,
        out_path: Union[str, Path],
        id: Optional[str] = "BF_FID",
        extra_footprints: Optional[gpd.GeoDataFrame] = None,
    ) -> gpd.GeoDataFrame:
        # Merge based on "id" column
        gdf = footprints[[id, "geometry"]].merge(points, on=id, how="outer")

        # Remove the building footprints without any object attached
        gdf = gdf.loc[~gdf["Object ID"].isna()]
        gdf["Object ID"] = gdf["Object ID"].astype(int)

        # Get columns that will be used
        strings = ["Primary Object Type"] + [
            col for col in gdf.columns if "Aggregation Label:" in col
        ]

        for col in strings:
            gdf[col] = gdf[col].astype(str)

        depths = []

        # Get type of run
        if "Total Damage" in gdf.columns:
            run_type = "event"
            # If event save inundation depth
            depths = depths + [col for col in gdf.columns if "Inundation Depth" in col]
            # And all type of damages
            dmgs = [
                col
                for col in gdf.columns
                if "Damage:" in col and "Max Potential" not in col
            ]
            dmgs.append("Total Damage")
        elif "Risk (EAD)" in gdf.columns:
            run_type = "risk"
            # For risk only save total damage per return period and EAD
            dmgs = [col for col in gdf.columns if "Total Damage" in col]
            dmgs.append("Risk (EAD)")
        else:
            raise ValueError(
                "The is no 'Total Damage' or 'Risk (EAD)' column in the results."
            )
        pot_dmgs = [col for col in gdf.columns if "Max Potential Damage" in col]
        dmgs = pot_dmgs + dmgs
        # Aggregate objects with the same "id"
        count = np.unique(gdf[id], return_counts=True)
        multiple_bffid = count[0][count[1] > 1][:-1]

        # First, combine the Primary Object Type and Object ID
        bffid_object_mapping = {}
        bffid_objectid_mapping = {}
        for bffid in multiple_bffid:
            all_objects = gdf.loc[gdf[id] == bffid, "Primary Object Type"].to_numpy()
            all_object_ids = gdf.loc[gdf[id] == bffid, "Object ID"].to_numpy()
            bffid_object_mapping.update(
                {bffid: "_".join(PointsToFootprints._mode(all_objects))}
            )
            bffid_objectid_mapping.update(
                {bffid: "_".join([str(x) for x in all_object_ids])}
            )
        gdf.loc[gdf[id].isin(multiple_bffid), "Primary Object Type"] = gdf[id].map(
            bffid_object_mapping
        )
        gdf.loc[gdf[id].isin(multiple_bffid), "Object ID"] = gdf[id].map(
            bffid_objectid_mapping
        )

        # Aggregated results using different functions based on type of output
        mapping = {}
        for name in strings:
            mapping[name] = pd.Series.mode
        for name in depths:
            mapping[name] = "mean"
        for name in dmgs:
            mapping[name] = "sum"

        agg_cols = strings + depths + dmgs

        df_groupby = (
            gdf.loc[gdf[id].isin(multiple_bffid), [id] + agg_cols]
            .groupby(id)
            .agg(mapping)
        )

        # Replace values in footprints file
        for agg_col in agg_cols:
            bffid_aggcol_mapping = dict(zip(df_groupby.index, df_groupby[agg_col]))
            gdf.loc[gdf[id].isin(multiple_bffid), agg_col] = gdf[id].map(
                bffid_aggcol_mapping
            )

        # Drop duplicates
        gdf = gdf.drop_duplicates(subset=[id])
        gdf = gdf.reset_index(drop=True)
        gdf = gdf[["Object ID", "geometry"] + agg_cols]

        for col in strings:
            for ind, val in enumerate(gdf[col]):
                if isinstance(val, np.ndarray):
                    gdf.loc[ind, col] = str(val[0])
        if extra_footprints is not None:
            extra_footprints = extra_footprints.to_crs(gdf.crs)
            # Merge based on "Object ID" column
            extra_footprints = extra_footprints[["Object ID", "geometry"]].merge(points, on="Object ID", how="left")[["Object ID", "geometry"]+agg_cols]
            gdf = pd.concat([gdf, extra_footprints], axis=0)
        
        # Calculate normalized damages per type
        value_cols = gdf.columns[gdf.columns.str.startswith('Max Potential Damage:')].tolist()
        
        # Only for event type calculate % damage per type
        if run_type == "event":
            dmg_cols = gdf.columns[gdf.columns.str.startswith('Damage:')].tolist()
            # Do per type
            for dmg_col in dmg_cols:
                new_name = dmg_col + " %"
                name = dmg_col.split("Damage: ")[1]
                gdf[new_name] = gdf[dmg_col] / gdf["Max Potential Damage: " + name] * 100
                gdf[new_name] = gdf[new_name].round(2)
            # Do total
            gdf["Total Damage %"] = gdf["Total Damage"] / gdf.loc[:, value_cols].sum(axis=1) * 100
            gdf["Total Damage %"] = gdf["Total Damage %"].round(2).fillna(0)
            
        # Calculate total normalized damage
        if run_type == "risk":
            tot_dmg_cols = gdf.columns[gdf.columns.str.startswith('Total Damage')].tolist()
            for tot_dmg_col in tot_dmg_cols:
                new_name = tot_dmg_col + " %"
                gdf[new_name] = gdf[tot_dmg_col] / gdf.loc[:, value_cols].sum(axis=1) * 100
                gdf[new_name] = gdf[new_name].round(2)
            gdf["Risk (EAD) %"] = gdf["Risk (EAD)"] / gdf.loc[:, value_cols].sum(axis=1) * 100
            gdf["Risk (EAD) %"] = gdf["Risk (EAD) %"].round(2).fillna(0)
        
        # Save file
        gdf.to_file(out_path, driver="GPKG")
        return gdf



from abc import ABC, abstractmethod


class IInfographicsParser(ABC):
    """Interface for creating the infographic"""

    @abstractmethod
    def get_infographics(self) -> str:
        """Get the infographic for a scenario

        Returns
        -------
        str
            The infographic for the scenario as a string in html format
        """
        pass

    @abstractmethod
    def write_infographics_to_file() -> str:
        """Write the infographic for a scenario to file

        Returns
        -------
        str
            The path to the infographic file
        """
        pass

    @abstractmethod
    def get_infographics_html() -> str:
        """Get the path to the infographic html file

        Returns
        -------
        str
            The path to the infographic html file
        """
        pass



import json
import logging
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Tuple, Union

import duckdb
import pandas as pd
import tomli

from fiat_toolbox.metrics_writer.fiat_metrics_interface import IMetricsFileWriter


# sql command struct
@dataclass
class sql_struct:
    name: str
    long_name: str
    show_in_metrics_table: bool
    description: str
    select: str
    filter: str
    groupby: str


class MetricsFileWriter(IMetricsFileWriter):
    """Class to parse metrics and write to a file."""

    def __init__(self, config_file: Union[str, Path]):
        """
        Initialize the class.

        Parameters
        ----------
        config_file : Union[str, Path]
            The path to the metrics file.
        """
        # Convert the path to a Path object
        if isinstance(config_file, str):
            config_file = Path(config_file)

        # Check whether the file exists
        if not os.path.exists(config_file):
            raise FileNotFoundError(f"Config file '{config_file}' not found.")

        self.config_file = config_file

    def _read_metrics_file(
        self, include_aggregates: bool
    ) -> Union[Dict[str, sql_struct], Dict[str, Dict[str, sql_struct]]]:
        """
        Read a metrics file and return a list of SQL commands.

        Parameters
        ----------
        include_aggregates : bool
            Whether to include aggregation labels in the metrics.

        Returns
        -------
        Union[Dict[str, sql_struct], Dict[str, Dict[str, sql_struct]]]
            A dictionary with the SQL commands.
        """

        # Read the metrics file
        _, extension = os.path.splitext(self.config_file)
        if extension == ".json":
            metrics = json.load(open(self.config_file, "r"))
        elif extension == ".toml":
            metrics = tomli.load(open(self.config_file, "rb"))
        else:
            raise ValueError(
                f"Config file '{self.config_file}' has an invalid extension. Only .json and .toml are supported."
            )

        # Create the sql commands dictionary
        sql_command_set = {}
        if include_aggregates:
            # Check whether the metrics file contains aggregation labels
            if "aggregateBy" not in metrics or len(metrics["aggregateBy"]) == 0:
                raise ValueError(
                    "No aggregation labels specified in the metrics file, but include_aggregates is set to True."
                )
            # Loop over the aggregation labels
            for aggregate in metrics["aggregateBy"]:
                aggregate_command = {}
                # Check whether the metrics file contains metrics
                if "queries" not in metrics or len(metrics["queries"]) == 0:
                    raise ValueError("No queries specified in the metrics file.")
                # Loop over the metrics
                for metric in metrics["queries"]:
                    # Check whether the metric contains all required fields
                    if all(
                        key in metrics
                        for key in [
                            "name",
                            "long_name",
                            "show_in_metrics_table",
                            "description",
                            "select",
                            "filter",
                        ]
                    ):
                        raise ValueError(
                            f"The metrics file for metric {metric['name']} does not contain all required fields."
                        )
                    # Create the sql command
                    sql_command = sql_struct(
                        name=metric["name"],
                        long_name=metric["long_name"],
                        show_in_metrics_table=metric["show_in_metrics_table"],
                        description=metric["description"],
                        select=metric["select"],
                        filter=metric["filter"],
                        groupby="`Aggregation Label: " + aggregate + "`",
                    )

                    # Check whether the metric name is already in the dictionary
                    if metric["name"] in aggregate_command:
                        raise ValueError(
                            f"Duplicate metric name {metric['name']} in metrics file."
                        )

                    # Add the sql command to the dictionary
                    aggregate_command[metric["name"]] = sql_command

                # Check whether the aggregation label is already in the dictionary
                if aggregate in sql_command_set:
                    raise ValueError(
                        f"Duplicate aggregation label {aggregate} in metrics file."
                    )

                # Add the sql command to the dictionary
                sql_command_set[aggregate] = aggregate_command
        else:
            # Check whether the metrics file contains metrics
            if "queries" not in metrics or len(metrics["queries"]) == 0:
                raise ValueError("No queries specified in the metrics file.")

            # Loop over the metrics
            for metric in metrics["queries"]:
                # Check whether the metric contains all required fields
                if all(
                    key in metrics
                    for key in [
                        "name",
                        "long_name",
                        "show_in_metrics_table",
                        "description",
                        "select",
                        "filter",
                    ]
                ):
                    raise ValueError(
                        f"The metrics file for metric {metric['name']} does not contain all required fields."
                    )

                # Create the sql command
                sql_command = sql_struct(
                    name=metric["name"],
                    long_name=metric["long_name"],
                    show_in_metrics_table=metric["show_in_metrics_table"],
                    description=metric["description"],
                    select=metric["select"],
                    filter=metric["filter"],
                    groupby="",
                )

                # Check whether the metric name is already in the dictionary
                if metric["name"] in sql_command_set:
                    raise ValueError(
                        f"Duplicate metric name {metric['name']} in metrics file."
                    )

                # Add the sql command to the dictionary
                sql_command_set[metric["name"]] = sql_command

        # Return the sql commands dictionary
        return sql_command_set

    @staticmethod
    def _create_single_metric(
        df_results: pd.DataFrame, sql_command: sql_struct
    ) -> Union[Tuple[str, object], Tuple[str, Dict[str, object]]]:
        """
        Create a metrics table from the results dataframe based on an SQL command.

        Parameters
        ----------
        df_results : pd.DataFrame
            The results dataframe.
        sql_command : sql_struct
            The SQL command.

        Returns
        -------
        Union[Tuple[str, object], Tuple[str, Dict[str, object]]]
            A tuple with the metric name and value or, in the case of a groupby statement,
            a tuple with the metric name and a dictionary with the groupby variables as keys
            and the metric as value.
        """

        # First add the the groupby variables to the query
        sql_query = "SELECT "
        if sql_command.groupby:
            sql_query += f"{sql_command.groupby}, "

        # Then add the select variables
        if not sql_command.select:
            raise ValueError(
                f"No select statement specified for metric {sql_command.name}."
            )
        sql_query += f"{sql_command.select} AS `{sql_command.name}` FROM df_results"

        # Then add the filter statement
        if sql_command.filter:
            sql_query += f" WHERE {sql_command.filter}"

        # Finally add the groupby statement
        if sql_command.groupby:
            sql_query += f" GROUP BY {sql_command.groupby}"

        # Execute the query. If the query is invalid, an error PandaSQLException will be raised
        sql_query = sql_query.replace("`", '"')
        result = duckdb.query(sql_query).df()

        # If the command contains a groupby statement, return a dictionary with the groupby variables as keys and the metric as value
        if sql_command.groupby:
            # Set the groupby variables as index
            labeled_result = result.set_index(sql_command.groupby.replace("`", ""))
            # Return the metric name and the dictionary
            return labeled_result.columns[0], dict(
                labeled_result[labeled_result.columns[0]]
            )
        # Otherwise return the metric name and the value
        return result.columns[0], result[result.columns[0]][0]

    @staticmethod
    def _create_metrics_dict(
        df_results: pd.DataFrame, sql_commands: Dict[str, sql_struct]
    ) -> Dict[str, object]:
        """
        Create a metrics table from the results dataframe based on a list of SQL commands.

        Parameters
        ----------
        df_results : pd.DataFrame
            The results dataframe.
        sql_commands : list[sql_struct]
            A list of SQL commands.

        Returns
        -------
        dict
            A dictionary with the metric names and values.
        """

        # Initialize the metrics dictionary
        df_metrics = {}

        # Run the sql commands one by one
        for name, command in sql_commands.items():
            # Create the metric (_create_single_metric is a static method, so no need to instantiate the class)
            _, value = MetricsFileWriter._create_single_metric(df_results, command)

            # Store the metric in the metrics dictionary using the metric name as key
            df_metrics[name] = value

        return df_metrics

    def _parse_metrics(
        self, df_results: pd.DataFrame, include_aggregates: bool
    ) -> Union[dict, Dict[str, dict]]:
        """
        Parse the metrics based on the config file and return a dictionary with the metrics.

        Parameters
        ----------
        df_results : pd.DataFrame
            The results dataframe.
        include_aggregates : bool
            Whether to include aggregation labels in the metrics.

        Returns
        -------
        Union[dict, List[dict]]
            A dictionary with the metrics or, in the case of multiple aggregation labels,
            a list of dictionaries.
        """

        # Read the metrics file
        sql_commands = self._read_metrics_file(include_aggregates)

        # Create the metrics dictionary
        if include_aggregates:
            metrics = {}
            # Loop over the aggregation labels
            for aggregate, commands in sql_commands.items():
                # Create the metrics dictionary for the current aggregation label (the _create_metrics_dict is a static method, so no need to instantiate the class)
                metrics[aggregate] = MetricsFileWriter._create_metrics_dict(
                    df_results, commands
                )
            return metrics
        else:
            # Create the metrics dictionary (the _create_metrics_dict is a static method, so no need to instantiate the class)
            return MetricsFileWriter._create_metrics_dict(df_results, sql_commands)

    @staticmethod
    def _write_metrics_file(
        metrics: Union[dict, Dict[str, dict]],
        config: Union[Dict[str, sql_struct], Dict[str, Dict[str, sql_struct]]],
        metrics_path: Path,
        write_aggregate: str = None,
        overwrite: bool = False,
    ) -> None:
        """
        Write a metrics dictionary to a metrics file.

        Parameters
        ----------
        metrics : Union[dict, List[dict]]
            A dictionary with the metrics or, in the case of multiple aggregation labels,
            a list of dictionaries.
        config : Union[Dict[str, sql_struct], Dict[str, Dict[str, sql_struct]]]
            A dictionary with the SQL commands.
        metrics_path : Path
            The path to where to store the metrics file.
        write_aggregate : str
            The name of the aggregation label to write to the metrics file (None for no aggregation label).
        overwrite : bool
            Whether to overwrite the existing metrics file if it already exists.

        Returns
        -------
        None
        """

        if write_aggregate:
            # Get the metrics for the current aggregation label
            aggregate_metrics = metrics[write_aggregate]

            # Find the names dynamically
            names = set()
            for value in aggregate_metrics.values():
                names.update(value.keys())

            # Update all empty metrics with 0
            for key, value in aggregate_metrics.items():
                if value == {}:
                    aggregate_metrics[key] = {name: 0 for name in names}
                    continue
                for name in names:
                    if name not in value:
                        aggregate_metrics[key][name] = 0

            # Create a dataframe from the metrics dictionary
            metricsFrame = (
                pd.DataFrame().from_dict(aggregate_metrics, orient="index").fillna(0)
            )

            # Add the long name to the dataframe
            metricsFrame.insert(
                0,
                "Long Name",
                [
                    config[write_aggregate][name].long_name
                    for name, _ in metricsFrame.iterrows()
                ],
            )

            # Add the metrics table selector to the dataframe
            metricsFrame.insert(
                0,
                "Show In Metrics Table",
                [
                    config[write_aggregate][name].show_in_metrics_table
                    for name, _ in metricsFrame.iterrows()
                ],
            )

            # Add the description to the dataframe
            metricsFrame.insert(
                0,
                "Description",
                [
                    config[write_aggregate][name].description
                    for name, _ in metricsFrame.iterrows()
                ],
            )

            # Check if the file already exists
            if os.path.exists(metrics_path):
                if overwrite:
                    logging.warning(
                        f"Metrics file '{metrics_path}' already exists. Overwriting..."
                    )
                    os.remove(metrics_path)
                else:
                    return

            # Transpose the dataframe
            metricsFrame = metricsFrame.transpose()

            # Write the metrics to a file
            if metrics_path.parent and not metrics_path.parent.exists():
                metrics_path.parent.mkdir(parents=True)
            metricsFrame.to_csv(metrics_path)
        else:
            # Create a dataframe from the metrics dictionary
            metricsFrame = (
                pd.DataFrame()
                .from_dict(metrics, orient="index", columns=["Value"])
                .fillna(0)
            )

            # Add the long name to the dataframe
            metricsFrame.insert(
                0,
                "Long Name",
                [config[name].long_name for name, _ in metricsFrame.iterrows()],
            )

            # Add the metrics table selector to the dataframe
            metricsFrame.insert(
                0,
                "Show In Metrics Table",
                [
                    config[name].show_in_metrics_table
                    for name, _ in metricsFrame.iterrows()
                ],
            )

            # Add the description to the dataframe
            metricsFrame.insert(
                0,
                "Description",
                [config[name].description for name, _ in metricsFrame.iterrows()],
            )

            # Check if the file already exists
            if os.path.exists(metrics_path):
                if overwrite:
                    logging.warning(
                        f"Metrics file '{metrics_path}' already exists. Overwriting..."
                    )
                    os.remove(metrics_path)
                else:
                    return

            # Write the metrics to a file
            if metrics_path.parent and not metrics_path.parent.exists():
                metrics_path.parent.mkdir(parents=True)
            metricsFrame.to_csv(metrics_path)

    def parse_metrics_to_file(
        self,
        df_results: pd.DataFrame,
        metrics_path: Union[str, Path],
        write_aggregate: str = None,
        overwrite: bool = False,
    ) -> Union[str, Dict[str, str]]:
        """
        Parse a metrics file and write the metrics to a file.

        Parameters
        ----------
        df_results : pd.DataFrame
            The results dataframe.
        metrics_path : Union[str, Path]
            The path to where to store the metrics file.
        write_aggregate : str
            The name of the aggregation label to write to the metrics file
            (None for no aggregation label, 'all' for all possible ones).
        overwrite : bool
            Whether to overwrite the existing metrics file if it already exists.

        Returns
        -------
        Union[str, Dict[str, str]]
            The path to the metrics file or a dictionary with the aggregation labels as keys
            and the paths to the metrics files as values.
        """

        # Convert the path to a Path object
        if isinstance(metrics_path, str):
            metrics_path = Path(metrics_path)

        # Check whether to include aggregation labels
        include_aggregates = True if write_aggregate else False

        # Read the metrics config file
        config = self._read_metrics_file(include_aggregates)

        # Parse the metrics
        metrics = self._parse_metrics(df_results, include_aggregates)

        # Write the metrics to a file
        if write_aggregate == "all":
            # Initialize the return dictionary
            return_files = {}
            for key in config.keys():
                # If using aggregation labels, add the aggregation label to the filename
                directory, filename = os.path.split(metrics_path)
                filename, extension = os.path.splitext(filename)
                new_filename = filename + "_" + key + extension
                new_path = Path(os.path.join(directory, new_filename))
                return_files[key] = new_path

                # Check if the file already exists and continue if it does
                if os.path.exists(new_path):
                    if not overwrite:
                        continue

                # Write the metrics to a file
                MetricsFileWriter._write_metrics_file(
                    metrics, config, new_path, write_aggregate=key, overwrite=overwrite
                )
        else:
            # Check if the file already exists and continue if it does
            if os.path.exists(metrics_path):
                if not overwrite:
                    return metrics_path

            # Write the metrics to a file
            MetricsFileWriter._write_metrics_file(
                metrics,
                config,
                metrics_path,
                write_aggregate=write_aggregate,
                overwrite=overwrite,
            )
            return_files = metrics_path

        return return_files



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional, Union

import geopandas as gpd
import pandas as pd

_FORMATS = ["geopackage", "shapefile", "GeoJSON"]


class IAggregationAreas(ABC):
    """Interface for writing an aggregation areas spatial file."""

    @abstractmethod
    def write_spatial_file(
        df_metrics: pd.DataFrame,
        gdf_aggr_areas: gpd.GeoDataFrame,
        out_path: Union[str, Path],
        id_name: Optional[str] = "name",
        file_format: Optional[str] = "geopackage",
    ) -> None:
        """Saves a geospatial file where the aggregation areas are join with metrics from a metric table.

        Parameters
        ----------
        df_metrics : pd.DataFrame
            dataframe containing the metrics
        gdf_aggr_areas : gpd.GeoDataFrame
            geodataframe with the aggregation areas (with an identifier column provided with argument "id_name")
        out_path : Union[str, Path]
            path where the geospatial file should be saved
        id_name : Optional[str], optional
            name of the identified column in gdf_aggr_areas to be used for the join, by default "name"
        file_format : Optional[str], optional
            file format of the output geospatial file, by default "geopackage"

        Raises
        ------
        ValueError
            If the given file format is not implemented.
        """
        pass


class AggregationAreas(IAggregationAreas):
    """Write an aggregation areas spatial file."""

    @staticmethod
    def _check_extension(out_path, ext):
        out_path = Path(out_path)
        if out_path.suffix != ext:
            raise ValueError(
                f"File extension given: '{out_path.suffix}' does not much the file format specified: {ext}."
            )

    @staticmethod
    def write_spatial_file(
        df_metrics: pd.DataFrame,
        gdf_aggr_areas: gpd.GeoDataFrame,
        out_path: Union[str, Path],
        id_name: Optional[str] = "name",
        file_format: Optional[str] = "geopackage",
    ) -> None:
        """Saves a geospatial file where the aggregation areas are join with metrics from a metric table.

        Parameters
        ----------
        df_metrics : pd.DataFrame
            dataframe containing the metrics
        gdf_aggr_areas : gpd.GeoDataFrame
            geodataframe with the aggregation areas (with an identifier column provided with argument "id_name")
        out_path : Union[str, Path]
            path where the geospatial file should be saved
        id_name : Optional[str], optional
            name of the identified column in gdf_aggr_areas to be used for the join, by default "name"
        file_format : Optional[str], optional
            file format of the output geospatial file, by default "geopackage"

        Raises
        ------
        ValueError
            If the given file format is not implemented.
        """
        # Get index as the first column
        index_name = df_metrics.columns[0]
        df_metrics = df_metrics.set_index(index_name)

        # Only keep metrics that are supposed to be in the metrics table
        metrics_to_keep = (
            df_metrics.loc["Show In Metrics Table", :]
            .map(lambda x: True if x == "True" else False)
            .astype(bool)
        )
        df = df_metrics.loc[:, metrics_to_keep]

        # Drop rows containing other variables
        df = df.drop(["Description", "Show In Metrics Table", "Long Name"])
        df = df.apply(pd.to_numeric)

        # Joins based on provided column name
        joined = gdf_aggr_areas.join(df, on=id_name)

        # Save file
        if file_format == "geopackage":
            AggregationAreas._check_extension(out_path, ".gpkg")
            joined.to_file(out_path, driver="GPKG")
        elif file_format == "shapefile":
            AggregationAreas._check_extension(out_path, ".shp")
            joined.to_file(out_path)
        elif file_format == "GeoJSON":
            AggregationAreas._check_extension(out_path, ".geojson")
            joined.to_file(out_path, driver="GeoJSON")
        else:
            raise ValueError(
                f"File format specified: {file_format} not in implemented formats: {*_FORMATS,}."
            )
